<?php
declare(strict_types=1);

namespace FiloBlu\ExtInventory\Model;

use Exception;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\DB\Adapter\AdapterInterface;
use Zend_Db;

/**
 * Class ReservationCleaner
 * @package FiloBlu\ExtInventory\Model
 */
class ReservationCleaner
{

    /**
     * @var ResourceConnection
     */
    private $resourceConnection;

    /**
     * @var AdapterInterface
     */
    private $connection;
    /**
     * @var string
     */
    private $reservationTable;
    /**
     * @var ReservationStatusFactory
     */
    private $reservationStatusFactory;

    /**
     * ReservationCleaner constructor.
     * @param ReservationStatusFactory $reservationStatusFactory
     * @param ResourceConnection $resourceConnection
     */
    public function __construct(ReservationStatusFactory $reservationStatusFactory, ResourceConnection $resourceConnection)
    {
        $this->resourceConnection = $resourceConnection;
        $this->reservationStatusFactory = $reservationStatusFactory;
    }

    /**
     * As described at https://devdocs.magento.com/guides/v2.3/inventory/reservations.html
     *
     * Reservation calculations
     *
     * Magento creates a reservation for each product when the following events occur:
     * - A customer or merchant places an order.
     * - A customer or merchant fully or partially cancels an order.
     * - The merchant creates a shipment for a physical product.
     * - The merchant creates an invoice for a virtual or downloadable product.
     * - The merchant issues a credit memo.
     *
     * Reservations are append-only operations, similar to a log of events.
     * The initial reservation is assigned a negative quantity value.
     * All subsequent reservations created while processing the order are positive values.
     * When the order is complete, the sum of all reservations for the product is 0.
     *
     * @param int $orderId
     * @return ReservationStatus
     */
    public function clean($orderId)
    {
        $connection = $this->getConnection();
        $reservationTable = $this->getReservationTable();
        $query = "SELECT * FROM {$reservationTable} WHERE `metadata` LIKE '%object_id\"%:%\"{$orderId}\"%'";
        $results = $connection->fetchAll($query, [], Zend_Db::FETCH_ASSOC);

        $ids = [];
        $totalQuantity = 0.0;
        /** @var ReservationStatus $reservationStatus */
        $reservationStatus = $this->reservationStatusFactory->create();
        $reservationStatus->setReservations($results);

        foreach ($results as $result) {
            $ids[] = $result['reservation_id'];
            $totalQuantity += (float)$result['quantity'];
        }

        if ($totalQuantity === 0.0) {
            $connection->delete($reservationTable, ['reservation_id IN (?)' => $ids]);
            return $reservationStatus->setStatus(ReservationStatus::STATUS_CLEANED);
        }

        if ($totalQuantity > 0) {
            return $reservationStatus->setStatus(ReservationStatus::STATUS_ERROR);
        }

        return $reservationStatus->setStatus(ReservationStatus::STATUS_PENDING);
    }

    /**
     * @return AdapterInterface
     */
    protected function getConnection()
    {
        if ($this->connection) {
            return $this->connection;
        }

        return ($this->connection = $this->resourceConnection->getConnection());
    }

    /**
     * @return string
     */
    protected function getReservationTable()
    {
        if ($this->reservationTable) {
            return $this->reservationTable;
        }

        return ($this->reservationTable = $this->resourceConnection->getTableName('inventory_reservation'));
    }

    /**
     * @return array
     * @throws Exception
     */
    public function getOrderIds()
    {
        $ids = [];
        $connection = $this->getConnection();
        $reservationTable = $this->getReservationTable();
        $query = null;
        try {
            $query = "SELECT DISTINCT(JSON_EXTRACT(metadata , '$.object_id')) AS `order_id` FROM `{$reservationTable}`";
            $results = $connection->fetchAll($query, [], Zend_Db::FETCH_ASSOC);
        } catch (Exception $exception) {
            // check for SQL Error [1305] [42000]: FUNCTION JSON_EXTRACT does not exist
            if ((int)$exception->getCode() === 42000) {
                $query = "SELECT DISTINCT(
SUBSTRING_INDEX(MID(REPLACE(REPLACE(REPLACE(`metadata`, ' ', ''), '\t', ''), '\n', ''), 
LOCATE( '\"object_id\"', REPLACE(REPLACE(REPLACE(`metadata`, ' ', ''), '\t', ''), '\n', '') ) + 13), '\"', 1)
) AS `order_id`
FROM `{$reservationTable}`";
                $results = $connection->fetchAll($query, [], Zend_Db::FETCH_ASSOC);
            } else {
                throw $exception;
            }
        }

        foreach ($results as $result) {
            $ids[] = (int)trim($result['order_id'], " \t\n\r\0\x0B\"");
        }

        return $ids;
    }
}
